print("import: 'dtaidistance'")
import dtaidistance

print("import: 'dtaidistance.clustering'")
import dtaidistance.clustering

print("import: 'dtaidistance.dtw'")
import dtaidistance.dtw

print("import: 'dtaidistance.dtw_cc'")
import dtaidistance.dtw_cc

print("import: 'dtaidistance.dtw_cc_numpy'")
import dtaidistance.dtw_cc_numpy

print("import: 'dtaidistance.dtw_cc_omp'")
import dtaidistance.dtw_cc_omp

print("import: 'dtaidistance.ed_cc'")
import dtaidistance.ed_cc

